/* Copyright	Massachusetts Institute of Technology	1987, 1988
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of M.I.T. not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  M.I.T. makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * M.I.T. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL M.I.T.
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */
/*
   Copyright 1994, Frederick W. Umminger.
   All rights reserved.
   */

extern "C" {
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <stdlib.h>
#include <stdio.h>
}


#include <iostream.h>
#include <fstream.h>
#include "JuggleP.h"

#ifdef DEBUG
#define debug(a) a
#else
#define debug(a)
#endif


static XtResource resources[] = {
#define offset(field) XtOffsetOf(JuggleRec, juggle.field)
    /* {name, class, type, size, offset, default_type, default_addr}, */
  { XtNforeground, XtCForeground, XtRPixel,
      sizeof(Pixel), offset(foreground), XtRPixel, XtDefaultForeground },
  { "number_hands", "Number_Hands", XtRInt,
      sizeof(int), offset(number_hands), XtRString, "2" },
  { "frames_per_second", "Frames_Per_Second", XtRFloat,
      sizeof(float), offset(frames_per_second), XtRString, "50" },
  { "throw_velocity", "Throw_Velocity", XtRFloat,
      sizeof(float), offset(throw_velocity), XtRString, "-128" },
   { "tv_ptr", "TV_Ptr", XtRPointer,
    sizeof(float*), offset(tv_ptr), XtRPointer, NULL },
   { "gravitational_force", "Gravitational_Force", XtRFloat,
    sizeof(float), offset(gravitational_force), XtRString, "2048" },
   { "gf_ptr", "GF_Ptr", XtRPointer,
    sizeof(float*), offset(gf_ptr), XtRPointer, NULL },
   { "pattern_filename", "Pattern_Filename", XtRString,
    sizeof(char*), offset(pattern_filename), 
    XtRString, "juggle_pattern" },
   { "pixmap_filename", "Pixmap_Filename", XtRString,
    sizeof(char*), offset(pixmap_filename), 
    XtRString, "ball.xpm" },
   { "width", "Width", XtRInt,
    sizeof(int), offset(width), 
    XtRString, "200" },
   { "height", "Height", XtRInt,
    sizeof(int), offset(height), 
    XtRString, "300" }
#undef offset
};

static XtActionsRec actions[] =
{
  /* {name, procedure}, */
  {"template", NULL}
};


static char translations[] = "";

static void Initialize(Widget wreq, Widget wnew, \
		       ArgList args, Cardinal *num_args);
static void Realize(Widget w, XtValueMask* v, XSetWindowAttributes* a);
static void Redisplay(Widget w, XEvent *event, Region region);
static void Resize(Widget w);
static Boolean SetValues(Widget current, Widget request, Widget wnew, ArgList args, Cardinal *num_args);


JuggleClassRec juggleClassRec = {
  { /* core fields */
    /* superclass		*/	(WidgetClass) &coreClassRec,
    /* class_name		*/	"Juggle",
    /* widget_size		*/	sizeof(JuggleRec),
    /* class_initialize		*/	NULL,
    /* class_part_initialize	*/	NULL,
    /* class_inited		*/	FALSE,
    /* initialize		*/	Initialize,
    /* initialize_hook		*/	NULL,
    /* realize			*/	Realize,
    /* actions			*/	actions,
    /* num_actions		*/	XtNumber(actions),
    /* resources		*/	resources,
    /* num_resources		*/	XtNumber(resources),
    /* xrm_class		*/	NULLQUARK,
    /* compress_motion		*/	TRUE,
    /* compress_exposure	*/	XtExposeCompressMultiple,
    /* compress_enterleave	*/	TRUE,
    /* visible_interest		*/	FALSE,
    /* destroy			*/	NULL,
    /* resize			*/	Resize,
    /* expose			*/	Redisplay,
    /* set_values		*/	SetValues,
    /* set_values_hook		*/	NULL,
    /* set_values_almost	*/	XtInheritSetValuesAlmost,
    /* get_values_hook		*/	NULL,
    /* accept_focus		*/	NULL,
    /* version			*/	XtVersion,
    /* callback_private		*/	NULL,
    /* tm_table			*/	translations,
    /* query_geometry		*/	XtInheritQueryGeometry,
    /* display_accelerator	*/	XtInheritDisplayAccelerator,
    /* extension		*/	NULL
  },
  { /* template fields */
    /* empty			*/	0
  }
};

WidgetClass juggleWidgetClass = (WidgetClass) &juggleClassRec;




static void Realize(Widget widget, XtValueMask* v, XSetWindowAttributes* a)
{
  JuggleWidget w = (JuggleWidget) widget;
  
  /* Call the core class ralize method */
  (*(&coreClassRec)->core_class.realize) (widget, v, a);
  
  /* Now do all of the initialization which requires knowing
     the window of the widget 
      */
  /* create the pixmaps for the balls */
  XpmAttributes xpm_attr;
  xpm_attr.valuemask = 0;
  if (
      XpmReadFileToPixmap(XtDisplay(w),
			  XtWindow(w),
			  w->juggle.pixmap_filename,
			  &w->juggle.ball_pixmap, 
			  &w->juggle.ball_shapemask, 
			  &xpm_attr
			  )
      != XpmSuccess
      ){
    cerr << "Pixmap read failed somehow.\n";
    exit(-1);
  };
  XpmFreeAttributes(&xpm_attr);
  w->juggle.ball_bitmap_width = xpm_attr.width;
  w->juggle.ball_bitmap_height = xpm_attr.height;
  w->juggle.ball_x_hotspot = xpm_attr.x_hotspot;
  w->juggle.ball_y_hotspot = xpm_attr.y_hotspot;
  
  w->juggle.buffer =   
    XCreatePixmap(XtDisplay(w),
		  XtWindow(w),
		  w->core.width,
		  w->core.height,
		  w->core.depth
		  );

   /* set up the GCs for these pixmaps */ 
  
  /* set up the GCs */
 XGCValues gcv;
  
  gcv.foreground = w->core.background_pixel;
  gcv.background = w->core.background_pixel;
  w->juggle.draw_gc = XCreateGC(XtDisplay(w), w->juggle.buffer, GCForeground | GCBackground, &gcv);
     XSetClipMask(XtDisplay(w), 
		  w->juggle.draw_gc, 
		  w->juggle.ball_shapemask);
   
  gcv.foreground = w->core.background_pixel;
  gcv.background = w->core.background_pixel;
  w->juggle.copy_gc = XtGetGC((Widget) w, GCForeground | GCBackground, &gcv);
  
};  



static void Redisplay(Widget widget, XEvent *event, Region region)
{
  
  JuggleWidget w = (JuggleWidget) widget;

};





static void Resize(Widget widget){
  JuggleWidget w = (JuggleWidget) widget;
  
  Pix p = w->juggle.hand_list_ptr->first();
  double xi = w->core.width/(w->juggle.number_hands+1);
  int j;
  for( j = 1 ; j <= w->juggle.number_hands ; j++){
    w->juggle.hand_list_ptr->operator()(p).x =  xi * j;
    w->juggle.hand_list_ptr->operator()(p).y =  (double) w->core.height - 40;
    w->juggle.hand_list_ptr->next(p);
  };

  XFreePixmap(XtDisplay(w),w->juggle.buffer);
  w->juggle.buffer =   
    XCreatePixmap(XtDisplay(w),
		  XtWindow(w),
		  w->core.width,
		  w->core.height,
		  w->core.depth
		  );

};







void ball_render( JuggleWidget w){
   DLList<Ball*> ball = *(w->juggle.ball_list_ptr); 

   Widget wnew = (Widget) w;
   Pix p = ball.first();
   while(p != (Pix) 0){
     /* draw into buffer */
     int x, y;
     x = (int) (ball(p)->x - w->juggle.ball_x_hotspot);
     y =  (int) (ball(p)->y - w->juggle.ball_y_hotspot);
     XSetClipOrigin(XtDisplay(w), w->juggle.draw_gc, \
		    x, y);
     XCopyArea(XtDisplay(w), 
		w->juggle.ball_pixmap, 
	       w->juggle.buffer, 
	       w->juggle.draw_gc,
	       0,0, 
		w->juggle.ball_bitmap_width, 
	       w->juggle.ball_bitmap_height,
	       x, y
	       );
     ball.next(p);
   };
 };




static void step(JuggleWidget w)
{
  Ball* ball_ptr;

  Pix p = w->juggle.ball_list_ptr->first();
  while(p != (Pix) 0){
    ball_ptr = w->juggle.ball_list_ptr->operator()(p);
       ball_ptr->vy = ball_ptr->vy 
      + w->juggle.gravitational_force/
	(
	 (w->juggle.frames_per_second)*(w->juggle.frames_per_second)
	 );
     ball_ptr->step();
    if(ball_ptr->ref_count == 0){
      if(ball_ptr->y > w->core.height + w->juggle.ball_bitmap_height){
	/* We don't need to erase the ball because it
	   is offscreen */
	delete(ball_ptr);
	w->juggle.ball_list_ptr->del(p);
      };
    };
    w->juggle.ball_list_ptr->next(p);
  };
};




 void Ball::step(){
   ox = x;
   oy = y;
   x = x + vx;
   y = y + vy;
 };


Ball::Ball(){
  ref_count = 1;
  x = 0;
  y = 0;
  vx= 0 ;
  vy = 0;
  ox= 0 ;
  oy = 0;
};



static void load_throwlist(JuggleWidget w){   
   /* 
     This currently takes a raw pattern.
     */

   int i;
   ifstream juggle_pattern_file(w->juggle.pattern_filename);

   w->juggle.juggle_pattern->clear();

   if(!juggle_pattern_file){
     cerr << "ERROR [Juggler::load_throwlist] : " \
       << "Cannot open juggle pattern file.\n";
   };
   while(juggle_pattern_file >> i){
     w->juggle.juggle_pattern->append(i);
   };
   if(w->juggle.juggle_pattern->empty()){
     cerr << "ERROR [Juggler::load_throwlist] : " \
       << "Juggle pattern file invalid!\n";
   };

 };





static void juggle_throw(JuggleWidget w){
  static Pix p = (Pix) 0;
  int throw_height;
  Ball* current_ball;
  double throw_time;

  
  /* height */ 
  {  
    if(p == (Pix) 0){
      load_throwlist(w);
      p = w->juggle.juggle_pattern->first();
      if (p == (Pix) 0){
	return;
      };
    };
    throw_height = w->juggle.juggle_pattern->operator()(p);
    w->juggle.juggle_pattern->next(p);
  };
  /* current_ball */
  {
    current_ball = w->juggle.state_machine->remove_front();
    
    if(current_ball == (Ball*) NULL){
      Ball* b = new(Ball);
      w->juggle.ball_list_ptr->append(b);
      current_ball = w->juggle.ball_list_ptr->rear();
    };
  };

  /* current_hand */
  {
    if(w->juggle.current_hand == 0){
      w->juggle.current_hand = w->juggle.hand_list_ptr->first();
      if(w->juggle.current_hand == 0){
	return; 
      };
    };
  };

  
  /* target_hand() */
    Pix h2 = w->juggle.current_hand;
  {
    int i;
    for(i = 0 ; i < throw_height; i++){
      w->juggle.hand_list_ptr->next(h2);
      if(h2 == (Pix) 0){
	h2 = w->juggle.hand_list_ptr->first();
      };
    };
  };
	  
  /* set_ball_state */
  {
    current_ball->x = 
      w->juggle.hand_list_ptr->operator()(w->juggle.current_hand).x;
    current_ball->y = 
      w->juggle.hand_list_ptr->operator()(w->juggle.current_hand).y;
    
    throw_time = (
		  (-2)*
		  ((w->juggle.throw_velocity)/
		   (w->juggle.gravitational_force))*
		  (throw_height)
		  );
    
    
    current_ball->vx = (1/throw_time)*
      ( 
     w->juggle.hand_list_ptr->operator()(h2).x - 
       (w->juggle.hand_list_ptr->operator()(w->juggle.current_hand).x) 
       )
	/(w->juggle.frames_per_second);
    
    current_ball->vy = throw_height* 
      (w->juggle.throw_velocity)/(w->juggle.frames_per_second);
  };
  
   /* state_machine_throw() */
  {
    Pix q,r;
    int i;
    q = w->juggle.state_machine->first();
    for(i = 1; i < throw_height; i++){
        r = q;
      w->juggle.state_machine->next(r);
        if(r == (Pix) 0){
	w->juggle.state_machine->append((Ball*) NULL);
	w->juggle.state_machine->next(q);
      }
      else{
	q = r;
      };
      };
      if(w->juggle.state_machine->operator()(q) == (Ball*) NULL){
      w->juggle.state_machine->operator()(q) = current_ball;
    }
    else{
	current_ball->ref_count--;
      };
  };

  w->juggle.hand_list_ptr->next(w->juggle.current_hand);

};








static void juggle_interrupt( Widget widget, XtIntervalId id)
{
  JuggleWidget w = (JuggleWidget) widget;
 
  /* Juggling */
  XFillRectangle(XtDisplay(w), 
	    w->juggle.buffer, 
	    w->juggle.copy_gc,
	    0,0, 
	    w->core.width, 
	    w->core.height
	    );

 step(w);
 ball_render(w);
 XCopyArea(XtDisplay(w), 
	    w->juggle.buffer, 
	    XtWindow(w), 
	    w->juggle.copy_gc,
	    0,0, 
	    w->core.width, 
	    w->core.height,
	    0,0
	    );
   if( w->juggle.counter < w->juggle.frames_per_throw){
     w->juggle.counter++;
   }
  else{
    w->juggle.counter = 1;
    juggle_throw(w);

  };

  XtAppAddTimeOut(XtWidgetToApplicationContext((Widget) w), 
		  w->juggle.time_per_frame, \
		  juggle_interrupt, w);
};






static void Initialize(Widget wreq, Widget wnew, \
		       ArgList args, Cardinal *num_args)
{
  JuggleWidget w = (JuggleWidget) wnew;

    /* Should check public values for sanity */
  
   /* private variables */

  /* Is this legit? */
  w->core.width = w->juggle.width;
  w->core.height = w->juggle.height;
  
   w->juggle.time_per_frame = (int) (1000/w->juggle.frames_per_second);
    w->juggle.frames_per_throw = (int)((-2)
				     *((w->juggle.throw_velocity)/
				       (w->juggle.gravitational_force))
				      *(w->juggle.frames_per_second)
				      );
  
  w->juggle.color_list_ptr = NULL;
  
  w->juggle.counter = 0;
  w->juggle.ball_list_ptr = new DLList<Ball*>;
  
  w->juggle.hand_list_ptr = new DLList<hand>;
  { /* initialize the list of hands */
    struct hand temp;
    w->juggle.hand_list_ptr->clear();
    temp.y = 250;
    int i;
    for(i = 1 ; i <= w->juggle.number_hands; i++){
      temp.x = (w->core.width/(w->juggle.number_hands+1))*i;
      w->juggle.hand_list_ptr->append(temp);
    };
    
    w->juggle.current_hand = 0;
    
    w->juggle.juggle_pattern = new SLList<int>;
    w->juggle.state_machine = new SLList<Ball*>;
      /* need 2 nulls due to shitty algorithm in throw()*/
    w->juggle.state_machine->append((Ball*) 0);
    w->juggle.state_machine->append((Ball*) 0);
    };
  XtAppAddTimeOut(XtWidgetToApplicationContext((Widget) w), \
		  1, \
		  juggle_interrupt, w);
};





static Boolean SetValues(Widget c, Widget r, Widget n,\
		       ArgList args, Cardinal *num_args)
{
/* WARNING: should be using XtIsRealized before XtWindow 
   Also, XSetForeground is WRONG, this is a shared GC and must be
   released and a new one grabbed.
 */

  JuggleWidget current = (JuggleWidget) c;
  JuggleWidget request = (JuggleWidget) r;
  JuggleWidget w = (JuggleWidget) n;

  Boolean redisplay = False;

   if( w->juggle.gf_ptr != NULL){
     w->juggle.gravitational_force = *(w->juggle.gf_ptr);
     w->juggle.gf_ptr = NULL;
   };

   if( w->juggle.tv_ptr != NULL){
     w->juggle.throw_velocity = *(w->juggle.tv_ptr);
     w->juggle.tv_ptr = NULL;
   };

   /* private variables */
   w->juggle.time_per_frame = (int) (1000/w->juggle.frames_per_second);
   w->juggle.frames_per_throw = (int)((-2)
				      *((w->juggle.throw_velocity)/
					(w->juggle.gravitational_force))
				      *(w->juggle.frames_per_second)
				      );
    return redisplay;
};









